;--------------------- printx macros -------------------------
;   copyright (C) 2008 Jim Giordano, free for all uses public, private, and commercial.
;      copies multiple strings to a buffer, optionally converting dwords
;      to text using dd keyword, or byte directly if immediate value < 256
.data
printxname  db "printx",0
printxstart dd 0    ; start of output buffer as defined by printx or printxf
printxptr   dd 0    ; next available spot for appending more text
printxsize  dd 0    ; current number of characters in the buffer
printxerr1  db "Error in printx: invalid input address",0
printxerr2  db "printx called without setting start address",0
printxerr3  db "Error in printx: invalid output address",0
printxeflg  db 0
.code
align 4
printxproc: ; proc ; the main printx procedure called by the printx macro
    ; all registers are saved by printxa macro
    ; edi = output address
    ; dh  = flag, 0=string, address in esi
    ;             1=signed dword in eax, convert to ascii
    ;             2=byte (no conversion, value in al)
    ;             3=unsigned dword in eax, convert to ascii
    ;             4=convert binary in eax to lower case hex ascii string (dl=bytes to convert)
    ;             5=convert binary in eax to upper case hex ascii string (dl=bytes to convert)
    ;             
    mov edi,printxptr    ;; where we will store the results
    .if edi ; has the user set an address?
        .if dh==1 || dh==3 ; flag to convert dword to text
            ; adapted from masm32 lib's dwtoa but can convert to both signed and unsigned values
            or eax,eax  ; test for zero and negative values
            .if Zero?
                mov word ptr [edi],30h  ; "0" and zero byte terminator
                inc edi     ; point at terminator
            .else
                .if Sign?    
                    .if dh==1    ; check signed
                        mov byte ptr [edi],'-'
                        neg eax
                        add edi, 1
                    .endif
                .endif
                mov ecx, 0CCCCCCCDh ; 8 * 1/10
                mov esi, edi    ; save start of string
                .repeat
                    mov ebx,eax
                    mul ecx
                    shr edx, 3
                    mov eax,edx
                    lea edx,[edx*4+edx]
                    add edx,edx
                    sub ebx,edx
                    add bl,'0'
                    mov [edi],bl
                    add edi, 1
                .until eax==0
                mov byte ptr [edi], 0   ; terminate the string
                mov ecx,edi
                .repeat                 ; We now have all the digits, but in reverse order.
                  sub ecx, 1            ; swap as needed
                  mov al, [esi]
                  mov ah, [ecx]
                  mov [ecx], al
                  mov [esi], ah
                  add esi, 1
                .until (esi >= ecx)
            .endif
        .elseif dh==2  ; flag to save a byte
            stosb
            mov byte ptr [edi],0
        .elseif dh==4 || dh==5 ; hex, dl=number of bytes
            .if dh==4
                mov dh,7   ; upper case
            .else
                mov dh,7+32   ; lower case
            .endif
            shl dl,1    ; two hex characters per byte
            mov cl,dl
            shl cl,2    ; * 4 bits per hex character
            ror eax,cl  ; move to upper end of word
            .repeat
                rol eax,4   ; get a nibble
                mov cl,al
                and ecx,0fh
                .if cl>9
                    add cl,dh    ; get up to an 'A' or 'a'
                 .endif
                 add cl,48      ; convert to ascii
                 mov [edi],cl
                 inc edi
                 dec dl
            .until zero?
            mov [edi],dl    ; terminator
        .else   ; a zbt string with address in esi
            mov printxeflg,0 ;Install per-thread error handler from http://www.winasm.net/forum/index.php?showtopic=2082
            assume fs:nothing ;MASM assumes the use of this register to be ERROR by default
            push offset PrintxExceptionHandler
            push fs:[0]
            mov fs:[0],esp
            .repeat
                lodsb
                stosb
            .until al==0
            dec edi
            PrintxErrRet:
            pop fs:[0]   ; Uninstall per-thread error handler
            add esp,4
            .if printxeflg
                mov eax,offset printxerr1
                jmp printxerrorexit
            .endif
        .endif
        mov printxptr,edi
        sub edi,printxstart
        mov printxsize,edi
    .else
        mov eax,offset printxerr2
        printxerrorexit:
        invoke MessageBox,0,eax,addr printxname,0
    .endif
	ret

PrintxExceptionHandler: ; PROC pExcept, pFrame, pContext, pDispatch
    mov eax,[esp+12]  ; pContext
    mov [eax].CONTEXT.regEip,offset PrintxErrRet
    mov eax,ExceptionContinueExecution
    mov printxeflg,1
    ret

CheckPrintxAddress:
    pusha
    mov printxptr,eax
    mov printxstart,eax
    mov printxsize,0

    mov printxeflg,0 ;Install per-thread error handler from http://www.winasm.net/forum/index.php?showtopic=2082
    assume fs:nothing ;MASM assumes the use of this register to be ERROR by default
    push offset CheckPrintxAddressExceptionHandler
    push fs:[0]
    mov fs:[0],esp
    
    mov dword ptr [eax],0

    CheckPrintxAddressErrRet:
    pop fs:[0]   ; Uninstall per-thread error handler
    add esp,4
    .if printxeflg
        mov eax,offset printxerr3
        invoke MessageBox,0,eax,addr printxname,0
        mov printxptr,0     ; don't use this address!
        mov printxstart,0
    .endif
    popa
    cmp printxstart,0   ; return flag
    ret

CheckPrintxAddressExceptionHandler:
    mov eax,[esp+12]  ; pContext
    mov [eax].CONTEXT.regEip,offset CheckPrintxAddressErrRet
    mov eax,ExceptionContinueExecution
    mov printxeflg,1
    ret

printx macro BufferAddr,Things:Vararg ;; copy all Things to BufferAddr
  ;; copys text strings surrounded by double or single quotes to buffer BufferAddr
  ;; an entry of the characters dd means the next value is a dword to be converted to text
  ;; printxa (printx append) may be called to append more text to current buffer
  ;; printxf is the function version of printx, which returns BufferAddr
  ;; printxaf is the function version of printxa, append to buffer, returns BufferAddr
    push eax
    if @SizeStr(BufferAddr) gt 5
        ifidni @SubStr(BufferAddr,1,5),<addr >
            printxbuffaddr textequ @SubStr(BufferAddr,6)
            if (opattr(printxbuffaddr)) eq 98
                lea eax,printxbuffaddr
            else
                mov eax,@CatStr(<offset >,printxbuffaddr)
            endif
        elseif @SizeStr(BufferAddr) gt 7
            ifidni @SubStr(BufferAddr,1,7),<offset >
                printxbuffaddr textequ @SubStr(BufferAddr,8)
                if (opattr(printxbuffaddr)) eq 98
                    lea eax,printxbuffaddr
                else
                    mov eax,BufferAddr
                endif
            else
                opxttr = opattr(BufferAddr)
                if opxttr eq 0
                    %echo printx: error in output buffer specification - BufferAddr
                    .err    
                    exitm
                endif   
                mov eax,BufferAddr
            endif
        else
            mov eax,BufferAddr
        endif
    else
        mov eax,BufferAddr        
    endif
    call CheckPrintxAddress ; go see if ok, and save if so
    pop eax
    .if !zero?
        printxa Things  ;; go append to buffer
    .endif   ; else error with buffer address 
endm
printxf macro BufferAddr,Things:Vararg ;; printx function returns buffer address
    printx BufferAddr,Things
    exitm <printxstart>
endm
printxaf macro ThingsToPrint:VARARG ;; printx append function, returns buffer address
    printxa ThingsToPrint
    exitm <printxstart>
endm
printxa macro ThingsToPrint:VARARG  ;; printx appends text to buffer starting at printxptr
    for arg,<ThingsToPrint>
        ;;xxxxxxxxxxxxxxxxxxxxxxxx textequ <&arg> ;; for debug
        dxflg=0     ;; assume string
        if @SizeStr(arg) eq 0
            goto DoneArg
        endif            
        pusha ;; save any registers that may be printed from later in this call
        if (@InStr(,<!"!'>,@SubStr(arg,1,1)))   ;; if single or double quote
            .data
               tmpoffset = $
               db arg,0
            .code
            mov esi,tmpoffset
            goto Doit
        endif
        if @SizeStr(arg) gt 3
            ifidni @SubStr(arg,1,3),<dd >
                dxflg = 1   ;; signed decimal
            else        ;; have to break up due to bug in masm.  jwasm ok.
              ifIDNi @SubStr(arg,1,3),<du >
                dxflg = 3   ;; unsigned decimal
              else
                 ifidni @SubStr(arg,1,3),<hx >
                    ifidn @SubStr(arg,1,1),<H>
                        dxflg = 4   ;; upper case hex
                    else
                        dxflg = 5   ;; lower case hex
                    endif
                 endif
              endif
            endif
        endif
        if dxflg
            bcnt=4    ;; assume 4 bytes
            opxttr = opattr(@SubStr(arg,4))
            ;;opxx textequ %opxttr
            ;;%echo opp= opxx &arg
            if opxttr eq 110000y ;; 30h a register
                if @SizeStr(@SubStr(arg,4)) eq 2
                    if dxflg eq 1
                        movsx eax,@SubStr(arg,4)    ;; signed
                    else ;; good for both du and hb, types 3 and 4
                        movzx eax,@SubStr(arg,4)
                    endif
                    ifidni @SubStr(arg,5),<h>  ;; type doesn't work on byte registers in masm
                        bcnt=1   ;; a byte (ah,bh,ch,dh)
                    else
                        ifidni @SubStr(arg,5),<l>
                            bcnt=1  ;; a byte (al,bl,cl,dl)
                        else
                            bcnt=2    ; was a word (ax,bx,si, etc.)
                        endif
                    endif
                else
                    mov eax,@SubStr(arg,4)
                endif
            elseif opxttr eq 100100y ;; 24h immediate value
                mov eax,@SubStr(arg,4)
            elseif opxttr eq 101010y or opxttr eq 100010y ;; 22h or 2Ah a variable name
                bcnt = type @SubStr(arg,4)
                if bcnt eq 2
                    if dxflg eq 1
                        movsx eax,word ptr @SubStr(arg,4)
                    else
                        movzx eax,word ptr @SubStr(arg,4)
                    endif
                elseif bcnt eq 1
                    if dxflg eq 1
                        movsx eax,byte ptr @SubStr(arg,4)
                    else
                        movzx eax,byte ptr @SubStr(arg,4)
                    endif
                else
                    mov eax,@SubStr(arg,4)
                    bcnt = 4
                endif
            else
                mov eax,@SubStr(arg,4)  ;; will produce an invalid address error is not valid
            endif
            if (dxflg eq 4) or (dxflg eq 5)
                mov dl,bcnt
            endif
        else

            if @SizeStr(arg) gt 3
                ifidni @SubStr(arg,1,3),<aa >
                    mov esi,@SubStr(arg,4)
                    goto Doit
                endif
            endif
            if @SizeStr(arg) gt 5
                ifidni @SubStr(arg,1,5),<addr >
                    lea esi,@SubStr(arg,6)
                    goto Doit
                endif
            endif          
            if (opattr(arg)) eq 100100y   ;; immediate value
                if (arg gt 0) and (arg lt 256)
                    mov al,arg
                    dxflg = 2
                else
                    mov eax,arg     ;; just print a large value
                    dxflg = 1
                endif
                goto Doit      ;; default to signed
            elseif (opattr(arg)) eq 98  ;; e.g. a proc parameter
                mov esi,arg
                goto Doit
            endif
            lea esi,[arg]
        endif
      :Doit
        mov dh,dxflg    ;; type for printxproc
        call printxproc
        popa
     :DoneArg
    endm
endm

msgx macro args:vararg  ;; show results of printx in message box
    pusha
    ifnb <args>
        printx args
    endif
    ifdef ProgName
        mov edx,offset ProgName
    else
        xor edx,edx
    endif
    call DoMsgx
    popa
endm

.code
DoMsgx:  ;call with eax=address of message, edx=address of ProgName or zero.  msgx saves all regs
    mov eax,printxstart
    .if eax
        invoke MessageBox,0,eax,edx,0
    .endif
    ret
    
ifndef m2m
m2m MACRO M1:req, M2:req
	push M2
	pop  M1
ENDM
endif
